local frameRateThresholds = {
    low = 50,      
    medium = 75,  
    high = 200,  
    max = 250      
}

local updateInterval = {
    dynamicMin = 5000,   
    dynamicMax = 20000  
}

local settings = {
    lodDistanceBias = 0.1, 
    pedLodBias = 0.1,       
    vehicleLodBias = 0.1,  
    particleLightMultiplier = 0.1, 
    skyQuality = 0.2,    
    pedDensity = 0.2,    
    vehicleDensity = 0.2,   
    reflectionQuality = 0.05, 
    shadowQuality = 0.05,  
    physicsQuality = 0.05, 
    bloomEnabled = false,   
    depthOfFieldEnabled = false, 
    volumetricLightingEnabled = false,  
    anisotropicFiltering = 1 
}

local frameGenerationEnabled = false

local lastInputTime = 0
local lastUpdateTime = 0
local fpsHistory = {}
local maxHistorySize = 10

local optimizing = true
local optimizationProgress = 0

local function aiPredictGraphicsAdjustment(fpsHistory)
    local sumFPS = 0
    for _, fps in ipairs(fpsHistory) do
        sumFPS = sumFPS + fps
    end
    local avgFPS = sumFPS / #fpsHistory

    if avgFPS < frameRateThresholds.low then
        return "reduce"
    elseif avgFPS < frameRateThresholds.medium then
        return "slightly_reduce"
    elseif avgFPS < frameRateThresholds.high then
        return "stable"
    else
        return "enhance"
    end
end

local function safeCall(func, ...)
    local success, result = pcall(func, ...)
    if not success then
        print("Error: " .. tostring(result))
        return false
    end
    return true
end

local function calculateFPS()
    local frameTime = GetFrameTime()
    return frameTime > 0 and (1 / frameTime) or 60
end

local function getSmoothedFPS()
    local fps = calculateFPS()
    table.insert(fpsHistory, fps)

    if #fpsHistory > maxHistorySize then
        table.remove(fpsHistory, 1)
    end

    local sum, weightSum = 0, 0
    for i, fpsValue in ipairs(fpsHistory) do
        local weight = #fpsHistory - i + 1
        sum = sum + (fpsValue * weight)
        weightSum = weightSum + weight
    end

    return sum / weightSum
end

local function checkForHack(fps)
    if fps > 500 or fps < 10 then
        print("[Anti-Hack] Abnormal FPS detected! Crashing game to prevent cheating.")
        os.exit()
    end
end

local function toggleFeature(feature, enable)
    settings[feature] = enable
    safeCall(SetVisualSettingFloat, feature, enable and 1.0 or 0.0)
    print(string.format("[Performance] %s %s.", feature, enable and "enabled" or "disabled"))
end

local function adjustGraphics(fps, action)
    local adjustmentFactor = (action == "reduce" and 0.1) or (action == "slightly_reduce" and 0.05) or 0.2
    local smoothFactor = 0.05  

    if action == "reduce" and fps < frameRateThresholds.low then
        adjustmentFactor = 0.3
    elseif action == "enhance" and fps > frameRateThresholds.max then
        adjustmentFactor = 0.15
    end

    for key, value in pairs(settings) do
        if type(value) == "number" then
            local newValue = math.max(math.min(value + (action == "reduce" and -adjustmentFactor or adjustmentFactor), 1.0), 0.1)
            settings[key] = settings[key] + (newValue - settings[key]) * smoothFactor  -- Gradual change
            safeCall(SetVisualSettingFloat, key, settings[key])
        elseif type(value) == "boolean" then
            toggleFeature(key, action ~= "reduce")
        end
    end

    print(string.format("[Performance] Graphics %s. Current FPS: %.1f", action, fps))
end

local function adjustGraphicsDynamically()
    local currentTime = GetGameTimer()
    local fps = getSmoothedFPS()

    checkForHack(fps) 

    local adjustedUpdateInterval = fps < 40 and updateInterval.dynamicMax or updateInterval.dynamicMin

    if currentTime - lastInputTime < adjustedUpdateInterval then return end
    lastInputTime = currentTime

    if currentTime - lastUpdateTime < updateInterval.dynamicMax then return end
    lastUpdateTime = currentTime

    local action = aiPredictGraphicsAdjustment(fpsHistory)
    adjustGraphics(fps, action)
    TriggerServerEvent('performance:updateGraphicsSettings', settings)
end

local function displayLoadingScreen()
    local width, height = GetScreenResolution()
    local progressBarWidth = 400
    local progressBarHeight = 40
    local barX = (width / 2) - (progressBarWidth / 2)
    local barY = (height / 2) - (progressBarHeight / 2)
    
    DrawRect(barX, barY, progressBarWidth, progressBarHeight, 0, 0, 0, 150) 

    local progressWidth = (progressBarWidth * optimizationProgress) / 100
    DrawRect(barX - (progressBarWidth / 2) + progressWidth / 2, barY, progressWidth, progressBarHeight, 0, 255, 0, 255) 

    SetTextFont(0)
    SetTextProportional(1)
    SetTextScale(0.5, 0.5)
    SetTextColour(255, 255, 255, 255)
    SetTextEntry("STRING")
    AddTextComponentString(string.format("Optimizing... %d%%", optimizationProgress))
    DrawText(barX + (progressBarWidth / 2) - 50, barY - 10)
end

local function performOptimization()
    optimizing = true
    local optimizationStartTime = GetGameTimer()

    for i = 1, 100 do
        optimizationProgress = i
        Citizen.Wait(100)
        if i == 100 then
            optimizing = false
        end
    end
end

Citizen.CreateThread(function()
    while true do
        Citizen.Wait(0)
        adjustGraphicsDynamically()

        if optimizing then
            displayLoadingScreen()
        end
    end
end)

AddEventHandler('playerSpawned', function()
    Citizen.Wait(0)
    adjustGraphicsDynamically()
end)

AddEventHandler('onClientMapStart', function()
    Citizen.Wait(0)
    adjustGraphicsDynamically()
    performOptimization()
end)

RegisterNetEvent('performance:applyGraphicsSettings')
AddEventHandler('performance:applyGraphicsSettings', function(newSettings)
    settings = newSettings
end)
